(** Module information *)

type topenwindowmode = WINDOW | FULLSCREEN
type tmousebutton = LEFT | RIGHT | MIDDLE
type twindowparam =
| OPENED 
| ACTIVE
| ICONIFIED
| ACCELERATED
| RED_BITS
| GREEN_BITS
| BLUE_BITS
| ALPHA_BITS
| DEPTH_BITS
| STENCIL_BITS
| REFRESH_RATE
| ACCUM_RED_BITS
| ACCUM_GREEN_BITS
| ACCUM_BLUE_BITS
| ACCUM_ALPHA_BITS
| AUX_BUFFERS
| STEREO

type tvidmode =
{
    vm_width:int;
    vm_height:int;
    vm_redbits:int;
    vm_greenbits:int;
    vm_bluebits:int;
}

type tjoyparam = PRESENT | AXES | BUTTONS

type tendis =
| MOUSE_CURSOR
| STICKY_KEYS
| STICKY_MOUSE_BUTTONS
| SYSTEM_KEYS
| KEY_REPEAT


let key_unknown = -1
let key_space = 32
let key_special = 128
let key_esc = 129
let key_f1 = 130
let key_f2 = 131
let key_f3 = 132
let key_f4 = 133
let key_f5 = 134
let key_f6 = 135
let key_f7 = 136
let key_f8 = 137
let key_f9 = 138
let key_f10 = 139
let key_f11 = 140
let key_f12 = 141
let key_up = 142
let key_down = 143
let key_left = 144
let key_right = 145
let key_lshift = 146
let key_rshift = 147
let key_lctrl = 148
let key_rctrl = 149
let key_lalt = 150
let key_ralt = 151
let key_tab = 152
let key_enter = 153
let key_backspace = 154
let key_insert = 155
let key_del = 156
let key_pageup = 157
let key_pagedown = 158
let key_home = 159
let key_end = 160
let key_kp_0 = 161
let key_kp_1 = 162
let key_kp_2 = 163
let key_kp_3 = 164
let key_kp_4 = 165
let key_kp_5 = 166
let key_kp_6 = 167
let key_kp_7 = 168
let key_kp_8 = 169
let key_kp_9 = 170
let key_kp_divide = 171
let key_kp_multiply = 172
let key_kp_subtract = 173
let key_kp_add = 174
let key_kp_decimal = 175
let key_kp_enter = 176


(** 
    Set alternate OpenGL library. default: "libGL.so.1"
*)
external set_gl_lib: string -> unit = "ml_glfwSetGLlib"

(** 
    Init Glfw. Must be called before any other function.
    @return false on failure
*)
external init : unit -> bool = "ml_glfwInit"

(** Terminate Glfw library. Must be called before exit. *)
external terminate: unit -> unit = "ml_glfwTerminate"

(** Get Glfw version. *)
external get_version: unit -> int * int * int = "ml_glfwGetVersion"

(**
    Open application window.
    
    {C {b open_window} {e width height redbits greenbits bluebits alphabits depthbits stencilbits mode}}
    @return false on failure
*)
external open_window: 
    int -> int -> 
    int -> int -> int -> 
    int -> int -> int -> 
    topenwindowmode -> bool = "byte_ml_glfwOpenWindow" "ml_glfwOpenWindow"

(** Set additional window information *)
external open_window_hint: twindowparam -> int -> unit = "ml_glfwOpenWindowHint"

(** Close application window *)
external close_window: unit -> unit = "ml_glfwCloseWindow"

(** Set window title *)
external set_window_title: string -> unit = "ml_glfwSetWindowTitle"

(** Get window size *)
external get_window_size: unit -> int * int = "ml_glfwGetWindowSize"

(** Set window size *)
external set_window_size: int -> int -> unit = "ml_glfwSetWindowSize"

(** Set window screen position *)
external set_window_pos: int -> int -> unit = "ml_glfwSetWindowPos"

(** Iconify window *)
external iconify_window: unit -> unit = "ml_glfwIconifyWindow"

(** Restore window *)
external restore_window: unit -> unit = "ml_glfwRestoreWindow"

(** Swap GL buffers *)
external swap_buffers: unit -> unit = "ml_glfwSwapBuffers"

(** 
    Minimum number of monitor vertical retraces between 
    each buffer swap performed by swap_buffers. 
    If interval is zero, buffer swaps will not be synchronized to the 
    vertical refresh of the monitor (aka 'VSync off').
*)
external swap_interval: int -> unit = "ml_glfwSwapInterval"

(** Get window parameter *)
external get_window_param: twindowparam -> int = "ml_glfwGetWindowParam"

external proto_swsc: unit -> unit = "ml_glfwProtoswsc"
(** set function to be called each time window size chaged *)
let set_window_size_cb (f:int -> int -> unit) =
    Callback.register "ml_setwindowsizecallback" f;
    proto_swsc ()

external proto_skc: unit -> unit = "ml_glfwProtoskc"
(** set function to be called each time key pressed/released *)
let set_key_cb (f: int -> bool -> unit) =
    Callback.register "ml_setkeycallback" f;
    proto_skc ()

external proto_mpc: unit -> unit = "ml_glfwProtompc"
(** set function to be called each time mouse position changed *)
let set_mouse_pos_cb (f: int -> int -> unit) =
    Callback.register "ml_setmouseposcallback" f;
    proto_mpc ()

external proto_mbc: unit -> unit = "ml_glfwProtombc"
(** set function to be called each time mouse button pressed/released *)
let set_mouse_button_cb (f: tmousebutton -> bool -> unit) =
    Callback.register "ml_setmousebuttoncallback" f;
    proto_mbc ()

external proto_mwc: unit -> unit = "ml_glfwProtomwc"
(** set function to be called each time mouse wheel moved *)
let set_mouse_wheel_cb (f: int -> unit) =
    Callback.register "ml_setmousewheelcallback" f;
    proto_mwc ()
    
(** get possible video modes *)
external get_video_modes: unit -> tvidmode array = "ml_glfwGetVideoModes"
(** get current video mode *)
external get_desktop_mode: unit -> tvidmode = "ml_glfwGetDesktopMode"

(** poll events *)
external poll_events: unit -> unit = "ml_glfwPollEvents"
(** get key state *)
external get_key: int -> bool = "ml_glfwGetKey"
(** get mouse button state *)
external get_mouse_button: tmousebutton -> bool = "ml_glfwGetMouseButton"
(** get mouse position *)
external get_mouse_pos: unit -> int * int = "ml_glfwGetMousePos"
(** get mouse wheel state *)
external get_mouse_wheel: unit -> int = "ml_glfwGetMouseWheel"
(** set mouse position *)
external set_mouse_pos: int -> int -> unit = "ml_glfwSetMousePos"
(** set mouse wheel state *)
external set_mouse_wheel: int -> unit = "ml_glfwSetMouseWheel"

(** get joystick param *)
external get_joystick_param: int -> tjoyparam -> int = "ml_glfwGetJoystickParam"
(** get joystick positions *)
external get_joystick_pos: int -> int -> float array = "ml_glfwGetJoystickPos"
(** get joystick buttons *)
external get_joystick_buttons: int -> int -> bool array = "ml_glfwGetJoystickButtons"

(** enable something *)
external enable: tendis -> unit = "ml_glfwEnable"
(** disable something *)
external disable: tendis -> unit = "ml_glfwDisable"
