//========================================================================
// GLFW - An OpenGL framework
// File:     window.c
// Platform: Windows
// Version:  2.3
// Date:     2002.11.23
// Author:   Marcus Geelnard (marcus.geelnard@home.se)
// WWW:      http://hem.passagen.se/opengl/glfw/
//------------------------------------------------------------------------
// Copyright (c) 2002 Marcus Geelnard
//
// GLFW is provided "as-is", without any express or implied warranty,
// without even the implied warranty of fitness for a particular purpose.
// In no event will the author be held liable for any damage araising from
// the use of GLFW.
//
// Permission is granted to anyone to use GLFW for any purpose, including
// commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. You must not claim that you wrote GLFW.
// 2. If you modify any parts of GLFW, it must be clearly stated which
//    modifications have been made, and that those modifications are not
//    the work of the author of GLFW.
// 3. If you use GLFW in a product, an acknowledgment in the product
//    documentation would be appreciated, but is not required.
// 4. This notice may not be removed or altered from any source
//    distribution of GLFW.
//
// Marcus Geelnard
// marcus.geelnard@home.se
//========================================================================

#include "internal.h"



//************************************************************************
//****                  GLFW internal functions                       ****
//************************************************************************


//========================================================================
// _glfwMinMaxAnimations() - Enable/disable minimize/restore animations
//========================================================================

static int _glfwMinMaxAnimations( int enable )
{
    ANIMATIONINFO AI;
    int old_enable;

    // Get old animation setting
    AI.cbSize = sizeof( ANIMATIONINFO );
    SystemParametersInfo( SPI_GETANIMATION, 0, &AI, 0 );
    old_enable = AI.iMinAnimate;

    // If requested, change setting
    if( old_enable != enable )
    {
        AI.iMinAnimate = enable;
        SystemParametersInfo( SPI_SETANIMATION, 0, &AI, SPIF_SENDCHANGE );
    }

    return old_enable;
}


//========================================================================
// _glfwSetForegroundWindow() - Function for bringing a window into focus
// and placing it on top of the window z stack. Due to some nastiness
// with how Win98/ME/2k/XP handles SetForegroundWindow, we have to go
// through some really bizarre measures to achieve this (thanks again, MS,
// for making life so much easier)!
//========================================================================

static void _glfwSetForegroundWindow( HWND hWnd )
{
    int try_count = 0;
    int old_animate;

    // Try the standard approach first...
    BringWindowToTop( hWnd );
    SetForegroundWindow( hWnd );

    // If it worked, return now
    if( hWnd == GetForegroundWindow() )
    {
        // Try to modify the system settings (since this is the foreground
        // process, we are allowed to do this)
        SystemParametersInfo( SPI_SETFOREGROUNDLOCKTIMEOUT, 0, (LPVOID)0,
                              SPIF_SENDCHANGE );
        return;
    }

    // For other Windows versions than 95 & NT4.0, the standard approach
    // may not work, so if we failed we have to "trick" Windows into
    // making our window the foureground window: Iconify and restore
    // again. It is ugly, but it seems to work (we turn off those annoying
    // zoom animations to make it look a bit better at least).

    // Turn off minimize/restore animations
    old_animate = _glfwMinMaxAnimations( 0 );

    // We try this a few times, just to be on the safe side of things...
    do
    {
        // Iconify & restore
        ShowWindow( hWnd, SW_HIDE );
        ShowWindow( hWnd, SW_SHOWMINIMIZED );
        ShowWindow( hWnd, SW_SHOWNORMAL );

        // Try to get focus
        BringWindowToTop( hWnd );
        SetForegroundWindow( hWnd );

        // We do not want to keep going on forever, so we keep track of
        // how many times we tried
        try_count ++;
    }
    while( hWnd != GetForegroundWindow() && try_count <= 3 );

    // Restore the system minimize/restore animation setting
    (void) _glfwMinMaxAnimations( old_animate );

    // Try to modify the system settings (since this is now hopefully the
    // foreground process, we are probably allowed to do this)
    SystemParametersInfo( SPI_SETFOREGROUNDLOCKTIMEOUT, 0, (LPVOID)0,
                          SPIF_SENDCHANGE );
}


//========================================================================
// _glfwClearInput() - Clear all input events
//========================================================================

static void _glfwClearInput( void )
{
    int i;

    // Release all keyboard keys
    for( i = 0; i <= GLFW_KEY_LAST; i ++ )
    {
        _glfwInput.Key[ i ] = 0;
    }

    // Release all mouse buttons
    for( i = 0; i <= GLFW_MOUSE_BUTTON_LAST; i ++ )
    {
        _glfwInput.MouseButton[ i ] = 0;
    }

    // Set mouse position to (0,0)
    _glfwInput.MousePosX = 0;
    _glfwInput.MousePosY = 0;

    // Set mouse wheel position to 0
    _glfwInput.WheelPos = 0;

    // The default is to use non sticky keys and mouse buttons
    _glfwInput.StickyKeys = GL_FALSE;
    _glfwInput.StickyMouseButtons = GL_FALSE;

    // The default is to disable key repeat
    _glfwInput.KeyRepeat = GL_FALSE;

    // Make sure that the LastPressedShiftKey field holds a valid value
    _glfwInput.LastPressedShiftKey = GLFW_KEY_LSHIFT;
}


//========================================================================
// _glfwInputKey() - Register keyboard activity
//========================================================================

static void _glfwInputKey( int key, int action )
{
    int keyrepeat = 0;

    if( key >= 0 && key <= GLFW_KEY_LAST )
    {
        // Are we trying to release an already released key?
        if( action == GLFW_RELEASE && _glfwInput.Key[ key ] != 1 )
        {
            return;
        }

        // Register key action
        if( action == GLFW_RELEASE && _glfwInput.StickyKeys )
        {
            _glfwInput.Key[ key ] = 2;
        }
        else
        {
            keyrepeat = (_glfwInput.Key[ key ] == GLFW_PRESS) &&
                        (action == GLFW_PRESS);
            _glfwInput.Key[ key ] = (char) action;
        }

        // Call user callback function
        if( _glfwKeyCallback && (_glfwInput.KeyRepeat || !keyrepeat) )
        {
            _glfwKeyCallback( key, action );
        }
    }
}


//========================================================================
// _glfwInputMouseClick() - Register mouse button clicks
//========================================================================

static void _glfwInputMouseClick( int button, int action )
{
    if( button >= 0 && button <= GLFW_MOUSE_BUTTON_LAST )
    {
        // Register mouse button action
        if( action == GLFW_RELEASE && _glfwInput.StickyMouseButtons )
        {
            _glfwInput.MouseButton[ button ] = 2;
        }
        else
        {
            _glfwInput.MouseButton[ button ] = (char) action;
        }

        // Call user callback function
        if( _glfwMouseButtonCallback )
        {
            _glfwMouseButtonCallback( button, action );
        }
    }
}


//========================================================================
// _glfwTranslateKey() - Translates a windows key to an internal coding
//========================================================================

static int _glfwTranslateKey( unsigned long wParam, unsigned long lParam )
{
    int lshift_down, rshift_down;
    MSG next_msg;
    DWORD msg_time;

    // Check which key was pressed or released
    switch( wParam )
    {
        // The SHIFT keys require special handling
        case VK_SHIFT:
            // Under Win 9x/ME we have to guess LEFT/RIGHT shift by
            // checking the raw keycode. 0x36 should be the right shift
            // key on most keyboards.
            if( _glfwSys.WinVer < _GLFW_WIN_NT4 )
            {
                if( (lParam & 0x01ff0000) == 0x00360000 )
                {
                    return GLFW_KEY_RSHIFT;
                }
                return GLFW_KEY_LSHIFT;
            }
            // Under NT/2k/XP we can check the current state of the
            // LEFT/RIGHT shift keys to determine which one generated this
            // event.
            lshift_down = GetKeyState( VK_LSHIFT ) & 0x8000;
            rshift_down = GetKeyState( VK_RSHIFT ) & 0x8000;
            if( (lshift_down && _glfwInput.Key[ GLFW_KEY_LSHIFT ] != 1) ||
                (!lshift_down && _glfwInput.Key[ GLFW_KEY_LSHIFT ] == 1) )
            {
                _glfwInput.LastPressedShiftKey = GLFW_KEY_LSHIFT;
                return GLFW_KEY_LSHIFT;
            }
            if( (rshift_down && _glfwInput.Key[ GLFW_KEY_RSHIFT ] != 1) ||
                (!rshift_down && _glfwInput.Key[ GLFW_KEY_RSHIFT ] == 1) )
            {
                _glfwInput.LastPressedShiftKey = GLFW_KEY_RSHIFT;
                return GLFW_KEY_RSHIFT;
            }
            // No state change, so we probably have a key repeat. Bit 30
            // of lParam is 1 if the key was held down before (i.e. it is
            // a repeated event).
            if( lParam & 0x40000000 )
            {
                return _glfwInput.LastPressedShiftKey;
            }
            return GLFW_KEY_UNKNOWN;

        // The CTRL keys require special handling
        case VK_CONTROL:
            // Is this an extended key (i.e. right key)?
            if( lParam & 0x01000000 )
            {
                return GLFW_KEY_RCTRL;
            }
            // Here is a trick: "Alt Gr" sends LCTRL, then RALT. We only
            // want the RALT message, so we try to see if the next message
            // is a RALT message. In that case, this is a false LCTRL!
            msg_time = GetMessageTime();
            if( PeekMessage( &next_msg, NULL, 0, 0, PM_NOREMOVE ) )
            {
                if( next_msg.message == WM_KEYDOWN ||
                    next_msg.message == WM_SYSKEYDOWN )
                {
                    if( next_msg.wParam == VK_MENU &&
                        (next_msg.lParam & 0x01000000) &&
                        next_msg.time == msg_time )
                    {
                        // Next message is a RALT down message, which
                        // means that this is NOT a proper LCTRL message!
                        return GLFW_KEY_UNKNOWN;
                    }
                }
            }
            return GLFW_KEY_LCTRL;

        // The ALT keys require special handling
        case VK_MENU:
            // Is this an extended key (i.e. right key)?
            if( lParam & 0x01000000 )
            {
                return GLFW_KEY_RALT;
            }
            return GLFW_KEY_LALT;

        // The ENTER keys require special handling
        case VK_RETURN:
            // Is this an extended key (i.e. right key)?
            if( lParam & 0x01000000 )
            {
                return GLFW_KEY_KP_ENTER;
            }
            return GLFW_KEY_ENTER;

        // Special keys (non character keys)
        case VK_ESCAPE:        return GLFW_KEY_ESC;
        case VK_TAB:           return GLFW_KEY_TAB;
        case VK_BACK:          return GLFW_KEY_BACKSPACE;
        case VK_HOME:          return GLFW_KEY_HOME;
        case VK_END:           return GLFW_KEY_END;
        case VK_PRIOR:         return GLFW_KEY_PAGEUP;
        case VK_NEXT:          return GLFW_KEY_PAGEDOWN;
        case VK_INSERT:        return GLFW_KEY_INSERT;
        case VK_DELETE:        return GLFW_KEY_DEL;
        case VK_LEFT:          return GLFW_KEY_LEFT;
        case VK_UP:            return GLFW_KEY_UP;
        case VK_RIGHT:         return GLFW_KEY_RIGHT;
        case VK_DOWN:          return GLFW_KEY_DOWN;
        case VK_F1:            return GLFW_KEY_F1;
        case VK_F2:            return GLFW_KEY_F2;
        case VK_F3:            return GLFW_KEY_F3;
        case VK_F4:            return GLFW_KEY_F4;
        case VK_F5:            return GLFW_KEY_F5;
        case VK_F6:            return GLFW_KEY_F6;
        case VK_F7:            return GLFW_KEY_F7;
        case VK_F8:            return GLFW_KEY_F8;
        case VK_F9:            return GLFW_KEY_F9;
        case VK_F10:           return GLFW_KEY_F10;
        case VK_F11:           return GLFW_KEY_F11;
        case VK_F12:           return GLFW_KEY_F12;
        case VK_SPACE:         return (int)' ';

        // Numeric keypad
        case VK_NUMPAD0:       return GLFW_KEY_KP_0;
        case VK_NUMPAD1:       return GLFW_KEY_KP_1;
        case VK_NUMPAD2:       return GLFW_KEY_KP_2;
        case VK_NUMPAD3:       return GLFW_KEY_KP_3;
        case VK_NUMPAD4:       return GLFW_KEY_KP_4;
        case VK_NUMPAD5:       return GLFW_KEY_KP_5;
        case VK_NUMPAD6:       return GLFW_KEY_KP_6;
        case VK_NUMPAD7:       return GLFW_KEY_KP_7;
        case VK_NUMPAD8:       return GLFW_KEY_KP_8;
        case VK_NUMPAD9:       return GLFW_KEY_KP_9;
        case VK_DIVIDE:        return GLFW_KEY_KP_DIVIDE;
        case VK_MULTIPLY:      return GLFW_KEY_KP_MULTIPLY;
        case VK_SUBTRACT:      return GLFW_KEY_KP_SUBTRACT;
        case VK_ADD:           return GLFW_KEY_KP_ADD;
        case VK_DECIMAL:       return GLFW_KEY_KP_DECIMAL;

        // Non alpha numeric keys - UK/US keyboard translation
        case 220:              return (int)'`';
        case 187:              return (int)'-';
        case 219:              return (int)'=';
        case 221:              return (int)'[';
        case 186:              return (int)']';
        case 192:              return (int)';';
        case 222:              return (int)'\'';
        case 191:              return (int)'#';
        case 226:              return (int)'\\';
        case 188:              return (int)',';
        case 190:              return (int)'.';
        case 189:              return (int)'/';

        // The rest (should be alpha numeric keys)
        default:
            if( wParam >= 'a' && wParam <= 'z' )
            {
                wParam = wParam - 'a' + 'A';
            }
            if( wParam >= ' ' && wParam <= '~' )
            {
                return wParam;
            }
            return GLFW_KEY_UNKNOWN;
    }
}


//========================================================================
// _glfwWindowCallback() - Window callback function (handles window
// events)
//========================================================================

LRESULT CALLBACK _glfwWindowCallback( HWND hWnd, UINT uMsg,
     WPARAM wParam, LPARAM lParam )
{
    int WheelDelta, Iconified;

    // Handle certain window messages
    switch( uMsg )
    {
        // Window activate message? (iconification?)
        case WM_ACTIVATE:
            _glfwWin.Active = LOWORD(wParam) != WA_INACTIVE ? GL_TRUE :
                                                              GL_FALSE;
            Iconified = HIWORD(wParam) ? GL_TRUE : GL_FALSE;

            // Were we deactivated/iconified?
            if( (!_glfwWin.Active || Iconified) && !_glfwWin.Iconified )
            {
                // If we are in fullscreen mode we need to iconify
                if( _glfwWin.Fullscreen )
                {
                    // Do we need to manually iconify?
                    if( !Iconified )
                    {
                        // Minimize window
                        CloseWindow( _glfwWin.Wnd );

                        // The window is now iconified
                        Iconified = GL_TRUE;
                    }

                    // Change display settings to the desktop resolution
                    ChangeDisplaySettings( NULL, 0 );
                }

                // Unlock mouse
                if( !_glfwWin.OldMouseLockValid )
                {
                    _glfwWin.OldMouseLock = _glfwWin.MouseLock;
                    _glfwWin.OldMouseLockValid = GL_TRUE;
                    glfwEnable( GLFW_MOUSE_CURSOR );
                }
            }
            else if( _glfwWin.Active || !Iconified )
            {
                // If we are in fullscreen mode we need to maximize
                if( _glfwWin.Fullscreen && _glfwWin.Iconified )
                {
                    // Change display settings to the user selected mode
                    _glfwSetVideoModeMODE( _glfwWin.ModeID );

                    // Do we need to manually restore window?
                    if( Iconified )
                    {
                        // Restore window
                        OpenIcon( _glfwWin.Wnd );

                        // The window is no longer iconified
                        Iconified = GL_FALSE;

                        // Activate window
                        ShowWindow( hWnd, SW_SHOW );
                        _glfwSetForegroundWindow( _glfwWin.Wnd );
                        SetFocus( _glfwWin.Wnd );
                    }
                }

                // Lock mouse, if necessary
                if( _glfwWin.OldMouseLockValid && _glfwWin.OldMouseLock )
                {
                    glfwDisable( GLFW_MOUSE_CURSOR );
                }
                _glfwWin.OldMouseLockValid = GL_FALSE;
            }

            _glfwWin.Iconified = Iconified;
            return 0;

        // Intercept system commands (forbid certain actions/events)
        case WM_SYSCOMMAND:
            switch( wParam )
            {
                // Screensaver trying to start or monitor trying to enter
                // powersave?
                case SC_SCREENSAVE:
                case SC_MONITORPOWER:
                    if( _glfwWin.Fullscreen )
                    {
                        return 0;
                    }
                    else
                    {
                        break;
                    }

                // User trying to access application menu using ALT?
                case SC_KEYMENU:
                    return 0;
            }
            break;

        // Did we receive a close message?
        case WM_CLOSE:
            PostQuitMessage( 0 );
            return 0;

        // Is a key being pressed?
        case WM_KEYDOWN:
        case WM_SYSKEYDOWN:
            _glfwInputKey( _glfwTranslateKey( wParam, lParam ),
                           GLFW_PRESS );
            return 0;

        // Is a key being released?
        case WM_KEYUP:
        case WM_SYSKEYUP:
            // Special trick: release both shift keys on SHIFT up event
            if( wParam == VK_SHIFT )
            {
                _glfwInputKey( GLFW_KEY_LSHIFT, GLFW_RELEASE );
                _glfwInputKey( GLFW_KEY_RSHIFT, GLFW_RELEASE );
            }
            else
            {
                _glfwInputKey( _glfwTranslateKey( wParam, lParam ),
                               GLFW_RELEASE );
            }
            return 0;

        // Were any of the mouse-buttons pressed?
        case WM_LBUTTONDOWN:
            _glfwInputMouseClick( GLFW_MOUSE_BUTTON_LEFT, GLFW_PRESS );
            return 0;
        case WM_RBUTTONDOWN:
            _glfwInputMouseClick( GLFW_MOUSE_BUTTON_RIGHT, GLFW_PRESS );
            return 0;
        case WM_MBUTTONDOWN:
            _glfwInputMouseClick( GLFW_MOUSE_BUTTON_MIDDLE, GLFW_PRESS );
            return 0;

        // Were any of the mouse-buttons released?
        case WM_LBUTTONUP:
            _glfwInputMouseClick( GLFW_MOUSE_BUTTON_LEFT, GLFW_RELEASE );
            return 0;
        case WM_RBUTTONUP:
            _glfwInputMouseClick( GLFW_MOUSE_BUTTON_RIGHT, GLFW_RELEASE );
            return 0;
        case WM_MBUTTONUP:
            _glfwInputMouseClick( GLFW_MOUSE_BUTTON_MIDDLE, GLFW_RELEASE );
            return 0;

        // Did the mouse move?
        case WM_MOUSEMOVE:
            if( LOWORD(lParam) != _glfwInput.OldMouseX ||
                HIWORD(lParam) != _glfwInput.OldMouseY )
            {
                if( _glfwWin.MouseLock )
                {
                    _glfwInput.MousePosX += LOWORD(lParam) -
                                            _glfwInput.OldMouseX;
                    _glfwInput.MousePosY += HIWORD(lParam) -
                                            _glfwInput.OldMouseY;
                }
                else
                {
                    _glfwInput.MousePosX = LOWORD(lParam);
                    _glfwInput.MousePosY = HIWORD(lParam);
                }
                _glfwInput.OldMouseX = LOWORD(lParam);
                _glfwInput.OldMouseY = HIWORD(lParam);
                _glfwInput.MouseMoved = GL_TRUE;

                // Call user callback function
                if( _glfwMousePosCallback )
                {
                    _glfwMousePosCallback( _glfwInput.MousePosX,
                                           _glfwInput.MousePosY );
                }
            }
            return 0;

        // Mouse wheel action?
        case WM_MOUSEWHEEL:
            // WM_MOUSEWHEEL is not supported under Windows 95
            if( _glfwSys.WinVer != _GLFW_WIN_95 )
            {
                WheelDelta = (((int)wParam) >> 16) / WHEEL_DELTA;
                _glfwInput.WheelPos += WheelDelta;
                if( _glfwMouseWheelCallback )
                {
                    _glfwMouseWheelCallback( _glfwInput.WheelPos );
                }
                return 0;
            }
            break;

        // Resize the window?
        case WM_SIZE:
            _glfwWin.Width  = LOWORD(lParam);
            _glfwWin.Height = HIWORD(lParam);
            if( _glfwWindowSizeCallback )
            {
                _glfwWindowSizeCallback( LOWORD(lParam), HIWORD(lParam) );
            }
            return 0;

        default:
            break;
    }

    // Pass all unhandled messages to DefWindowProc
    return DefWindowProc( hWnd, uMsg, wParam, lParam );
}



//************************************************************************
//****                    GLFW user functions                         ****
//************************************************************************

//========================================================================
// glfwOpenWindow() - Here is where the window is created, and the OpenGL
// rendering context is created.
//========================================================================

GLFWAPI int GLFWAPIENTRY glfwOpenWindow( int width, int height,
    int redbits, int greenbits, int bluebits, int alphabits,
    int depthbits, int stencilbits, int mode )
{
    GLuint      PixelFormat;
    WNDCLASS    wc;
    DWORD       dwExStyle, dwStyle;
    RECT        WindowRect;
    PIXELFORMATDESCRIPTOR pfd;
    BYTE        AccumRedBits, AccumGreenBits, AccumBlueBits;
    BYTE        AccumAlphaBits, AuxBuffers, Stereo;

    // Is GLFW initialized?
    if( !_glfwInitialised || _glfwWin.Opened )
    {
        return GL_FALSE;
    }

    // Check input arguments
    if( mode != GLFW_WINDOW && mode != GLFW_FULLSCREEN )
    {
        return GL_FALSE;
    }

    // Clear GLFW state
    _glfwWin.DC                = NULL;
    _glfwWin.RC                = NULL;
    _glfwWin.Wnd               = NULL;
    _glfwWin.Instance          = NULL;
    _glfwWin.Active            = GL_TRUE;
    _glfwWin.Iconified         = GL_FALSE;
    _glfwWin.OldMouseLockValid = GL_FALSE;
    _glfwClearInput();

    // Unregister all callback functions
    _glfwWindowSizeCallback  = NULL;
    _glfwKeyCallback         = NULL;
    _glfwMousePosCallback    = NULL;
    _glfwMouseButtonCallback = NULL;
    _glfwMouseWheelCallback  = NULL;

    // Get window hints
    _glfwWin.RefreshRate = _glfwWinHints.RefreshRate;
    AccumRedBits         = (BYTE) _glfwWinHints.AccumRedBits;
    AccumGreenBits       = (BYTE) _glfwWinHints.AccumGreenBits;
    AccumBlueBits        = (BYTE) _glfwWinHints.AccumBlueBits;
    AccumAlphaBits       = (BYTE) _glfwWinHints.AccumAlphaBits;
    AuxBuffers           = (BYTE) _glfwWinHints.AuxBuffers;
    Stereo               = (BYTE) _glfwWinHints.Stereo;

    // Clear window hints
    _glfwWinHints.RefreshRate    = 0;
    _glfwWinHints.AccumRedBits   = 0;
    _glfwWinHints.AccumGreenBits = 0;
    _glfwWinHints.AccumBlueBits  = 0;
    _glfwWinHints.AccumAlphaBits = 0;
    _glfwWinHints.AuxBuffers     = 0;
    _glfwWinHints.Stereo         = 0;

    // Check width & height
    if( width > 0 && height <= 0 )
    {
        // Set the window aspect ratio to 4:3
        height = (width * 3) / 4;
    }
    else if( width <= 0 && height > 0 )
    {
        // Set the window aspect ratio to 4:3
        width = (height * 4) / 3;
    }
    else if( width <= 0 && height <= 0 )
    {
        // Default window size
        width  = 640;
        height = 480;
    }

    // Remeber window settings
    _glfwWin.Width      = width;
    _glfwWin.Height     = height;
    _glfwWin.Fullscreen = (mode == GLFW_FULLSCREEN ? 1 : 0);

    // Grab an instance for our window
    _glfwWin.Instance = GetModuleHandle( NULL );

    // Set window class parameters
    wc.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC; // Redraw on...
    wc.lpfnWndProc   = (WNDPROC)_glfwWindowCallback;  // Message handler
    wc.cbClsExtra    = 0;                             // No extra class data
    wc.cbWndExtra    = 0;                             // No extra window data
    wc.hInstance     = _glfwWin.Instance;             // Set instance
    wc.hIcon         = LoadIcon( NULL, IDI_WINLOGO ); // Load default icon
    wc.hCursor       = LoadCursor( NULL, IDC_ARROW ); // Load arrow pointer
    wc.hbrBackground = NULL;                          // No background
    wc.lpszMenuName  = NULL;                          // No menu
    wc.lpszClassName = "GLFW";                        // Set class name

    // Register the window class
    if( !RegisterClass( &wc ) )
    {
        glfwCloseWindow();
        return GL_FALSE;
    }

    // Do we want full-screen mode?
    if( _glfwWin.Fullscreen )
    {
        _glfwSetVideoMode( &_glfwWin.Width, &_glfwWin.Height,
                           redbits, greenbits, bluebits,
                           _glfwWin.RefreshRate );
    }

    // Set window style, depending on fullscreen mode
    if( _glfwWin.Fullscreen )
    {
        dwStyle = WS_POPUP;
        // Here's a trick for helping us getting window focus
        // (SetForegroundWindow doesn't work properly under
        // Win98/ME/2K/XP+)
        if( !(_glfwSys.WinVer == _GLFW_WIN_95 ||
              _glfwSys.WinVer == _GLFW_WIN_NT4) )
        {
            dwStyle = dwStyle | WS_MINIMIZE;
        }
        else
        {
            dwStyle = dwStyle | WS_VISIBLE;
        }
        dwExStyle = WS_EX_APPWINDOW;
    }
    else
    {
        dwStyle   = WS_OVERLAPPEDWINDOW | WS_CAPTION | WS_VISIBLE;
        dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
    }

    // Add required window styles
    dwStyle = dwStyle | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;

    // Set window size to true requested size (adjust for window borders)
    WindowRect.left   = (long)0;
    WindowRect.right  = (long)_glfwWin.Width;
    WindowRect.top    = (long)0;
    WindowRect.bottom = (long)_glfwWin.Height;
    AdjustWindowRectEx( &WindowRect, dwStyle, FALSE, dwExStyle );

    // Create window
    _glfwWin.Wnd = CreateWindowEx(
               dwExStyle,                          // Extended style
               "GLFW",                             // Class name
               "GLFW Window",                      // Window title
               dwStyle,                            // Defined window style
               0, 0,                               // Window position
               WindowRect.right-WindowRect.left,   // Calculate window width
               WindowRect.bottom-WindowRect.top,   // Calculate window height
               NULL,                               // No parent window
               NULL,                               // No menu
               _glfwWin.Instance,                  // Instance
               NULL );                             // Nothing to WM_CREATE
    if( !_glfwWin.Wnd )
    {
        glfwCloseWindow();
        return GL_FALSE;
    }

    // Get a device context
    _glfwWin.DC = GetDC( _glfwWin.Wnd );
    if( !_glfwWin.DC )
    {
        glfwCloseWindow();
        return GL_FALSE;
    }

    // Set required pixel format
    pfd.nSize           = sizeof(PIXELFORMATDESCRIPTOR);
    pfd.nVersion        = 1;
    pfd.dwFlags         = PFD_DRAW_TO_WINDOW | // Draw to window
                          PFD_SUPPORT_OPENGL | // Support OpenGL
                          PFD_DOUBLEBUFFER;    // Double buffered window
    pfd.iPixelType      = PFD_TYPE_RGBA;       // Request an RGBA format
    pfd.cColorBits      = 0;                   // Color bits ignored
    pfd.cRedBits        = (BYTE) redbits;      // Red bits
    pfd.cRedShift       = 0;                   // Red shift ignored
    pfd.cGreenBits      = (BYTE) greenbits;    // Green bits
    pfd.cGreenShift     = 0;                   // Green shift ignored
    pfd.cBlueBits       = (BYTE) bluebits;     // Blue bits
    pfd.cBlueShift      = 0;                   // Blue shift ignored
    pfd.cAlphaBits      = (BYTE) alphabits;    // Alpha bits
    pfd.cAlphaShift     = 0;                   // Alpha shift ignored
    pfd.cAccumBits      = 0;                   // Accum. bits ignored
    pfd.cAccumRedBits   = AccumRedBits;        // Accum. red bits
    pfd.cAccumGreenBits = AccumGreenBits;      // Accum. green bits
    pfd.cAccumBlueBits  = AccumBlueBits;       // Accum. blue bits
    pfd.cAccumAlphaBits = AccumAlphaBits;      // Accum. alpha bits
    pfd.cDepthBits      = (BYTE) depthbits;    // Depth buffer bits
    pfd.cStencilBits    = (BYTE) stencilbits;  // Stencil buffer bits
    pfd.cAuxBuffers     = AuxBuffers;          // No. of aux buffers
    pfd.iLayerType      = PFD_MAIN_PLANE;      // Drawing layer: main
    pfd.bReserved       = 0;                   // (reserved)
    pfd.dwLayerMask     = 0;                   // Ignored
    pfd.dwVisibleMask   = 0;                   // "
    pfd.dwDamageMask    = 0;                   // "
    if( depthbits <= 0 )
    {
        // We do not need a depth buffer
        pfd.dwFlags |= PFD_DEPTH_DONTCARE;
    }
    if( Stereo )
    {
        // Request a stereo mode
        pfd.dwFlags |= PFD_STEREO;
    }

    // Find a matching pixel format
    if( !(PixelFormat = _glfw_ChoosePixelFormat( _glfwWin.DC, &pfd )) )
    {
        glfwCloseWindow();
        return GL_FALSE;
    }

    // Set the pixel-format
    if( !_glfw_SetPixelFormat( _glfwWin.DC, PixelFormat, &pfd ) )
    {
        glfwCloseWindow();
        return GL_FALSE;
    }

    // Get a rendering context
    _glfwWin.RC = wglCreateContext(_glfwWin.DC);
    if( !_glfwWin.RC )
    {
        glfwCloseWindow();
        return GL_FALSE;
    }

    // Activate the OpenGL rendering context
    if( !wglMakeCurrent( _glfwWin.DC, _glfwWin.RC ) )
    {
        glfwCloseWindow();
        return GL_FALSE;
    }

    // Make sure that our window ends up on top of things (we call
    // ShowWindow twice here, since the first call may be ignored)
    if( dwStyle & WS_MINIMIZE )
    {
        // Turn off minimize/restore animations
        int old_animate = _glfwMinMaxAnimations( 0 );

        // The window was opened in minimized state, so now maximize it
        // (this is a trick to help us getting input focus)
        ShowWindow( _glfwWin.Wnd, SW_HIDE );
        ShowWindow( _glfwWin.Wnd, SW_HIDE );
        ShowWindow( _glfwWin.Wnd, SW_SHOWMINIMIZED );
        ShowWindow( _glfwWin.Wnd, SW_SHOWNORMAL );

        // Restore the system minimize/restore animation setting
        (void) _glfwMinMaxAnimations( old_animate );
    }
    else
    {
        // The window was opened in normal (visible) state
        ShowWindow( _glfwWin.Wnd, SW_SHOW );
        ShowWindow( _glfwWin.Wnd, SW_SHOW );
    }
    _glfwSetForegroundWindow( _glfwWin.Wnd );
    SetFocus( _glfwWin.Wnd );

    // Start by clearing the front buffer to black (avoid ugly desktop
    // remains in our OpenGL window)
    glClear( GL_COLOR_BUFFER_BIT );
    _glfw_SwapBuffers( _glfwWin.DC );

    // Initialize OpenGL extension: WGL_EXT_swap_control
    _wglSwapIntervalEXT = (WGLSWAPINTERVALEXT_T)
                          wglGetProcAddress("wglSwapIntervalEXT");

    // Clear OpenGL extension: GL_SGIS_generate_mipmap (will be
    // initialized in glfwLoadTexture2D)
    _glfwWin.Has_GL_SGIS_generate_mipmap = -1;

    // Flag that window is now opened
    _glfwWin.Opened = GL_TRUE;

    // If full-screen mode was requested, lock mouse (hide mouse cursor
    // and enable mouse input)
    if( mode == GLFW_FULLSCREEN )
    {
        glfwDisable( GLFW_MOUSE_CURSOR );
    }

    return GL_TRUE;
}


//========================================================================
// glfwOpenWindowHint() - Set hints for opening the window
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwOpenWindowHint( int target, int hint )
{
    // Is GLFW initialized?
    if( !_glfwInitialised )
    {
        return;
    }

    switch( target )
    {
        case GLFW_REFRESH_RATE:
            _glfwWinHints.RefreshRate = hint;
            break;
        case GLFW_ACCUM_RED_BITS:
            _glfwWinHints.AccumRedBits = hint;
            break;
        case GLFW_ACCUM_GREEN_BITS:
            _glfwWinHints.AccumGreenBits = hint;
            break;
        case GLFW_ACCUM_BLUE_BITS:
            _glfwWinHints.AccumBlueBits = hint;
            break;
        case GLFW_ACCUM_ALPHA_BITS:
            _glfwWinHints.AccumAlphaBits = hint;
            break;
        case GLFW_AUX_BUFFERS:
            _glfwWinHints.AuxBuffers = hint;
            break;
        case GLFW_STEREO:
            _glfwWinHints.Stereo = hint;
            break;
        default:
            break;
    }
}


//========================================================================
// glfwCloseWindow() - Properly kill the video display.
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwCloseWindow( void )
{
    // Is GLFW initialized?
    if( !_glfwInitialised )
    {
        return;
    }

    // Window is no longer opened
    _glfwWin.Opened = GL_FALSE;

    // Unregister all callback functions
    _glfwWindowSizeCallback  = NULL;
    _glfwKeyCallback         = NULL;
    _glfwMousePosCallback    = NULL;
    _glfwMouseButtonCallback = NULL;
    _glfwMouseWheelCallback  = NULL;

    // Show mouse pointer again (if hidden)
    glfwEnable( GLFW_MOUSE_CURSOR );

    // Inactivate OpenGL extension WGL_EXT_swap_control (activated by
    // glfwOpenWindow)
    _wglSwapIntervalEXT = (WGLSWAPINTERVALEXT_T) 0;

    // Do we have a rendering context?
    if( _glfwWin.RC )
    {
        // Release the DC and RC contexts
        wglMakeCurrent( NULL, NULL );

        // Delete the rendering context
        wglDeleteContext( _glfwWin.RC );
        _glfwWin.RC = NULL;
    }

    // Do we have a device context?
    if( _glfwWin.DC )
    {
        // Release the device context
        ReleaseDC( _glfwWin.Wnd, _glfwWin.DC );
        _glfwWin.DC = NULL;
    }

    // Do we have a window?
    if( _glfwWin.Wnd )
    {
        // Destroy the window
        DestroyWindow( _glfwWin.Wnd );
        _glfwWin.Wnd = NULL;
    }

    // Do we have an instance?
    if( _glfwWin.Instance )
    {
        // Unregister class
        UnregisterClass( "GLFW", _glfwWin.Instance );
        _glfwWin.Instance = NULL;
    }

    // Are we in fullscreen mode?
    if( _glfwWin.Fullscreen )
    {
        // Switch back to desktop resolution
        ChangeDisplaySettings( NULL, 0 );
    }
}


//========================================================================
// glfwSetWindowTitle() - Set the window title.
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwSetWindowTitle( const char *title )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Set window title
    (void) SetWindowText( _glfwWin.Wnd, title );
}


//========================================================================
// glfwGetWindowSize() - Get the window size.
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwGetWindowSize( int *width, int *height )
{
    if( width != NULL )
    {
        *width = _glfwWin.Width;
    }
    if( height != NULL )
    {
        *height = _glfwWin.Height;
    }
}


//========================================================================
// glfwSetWindowSize() - Set the window size.
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwSetWindowSize( int width, int height )
{
    int     bpp, mode = 0, refresh;
    int     sizechanged = GL_FALSE;
    GLint   drawbuffer;
    GLfloat clearcolor[4];

    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened || _glfwWin.Iconified )
    {
        return;
    }

    // If we are in fullscreen mode, get some info about the current mode
    if( _glfwWin.Fullscreen )
    {
        DEVMODE dm;
        int success;

        // Get current BPP settings
        dm.dmSize = sizeof( DEVMODE );
        success = EnumDisplaySettings( NULL, _glfwWin.ModeID, &dm );
        if( success )
        {
            // Get bpp
            bpp = dm.dmBitsPerPel;

            // Get closest match for target video mode
            refresh = _glfwWin.RefreshRate;
            mode = _glfwGetClosestVideoModeBPP( &width, &height, &bpp,
                                                &refresh );
        }
        else
        {
            mode = _glfwWin.ModeID;
        }
    }

    // Change window size before changing fullscreen mode?
    if( _glfwWin.Fullscreen && (width > _glfwWin.Width) )
    {
        SetWindowPos( _glfwWin.Wnd, HWND_TOP, 0, 0, width, height,
                      SWP_NOOWNERZORDER | SWP_NOMOVE | SWP_NOZORDER );
        sizechanged = GL_TRUE;
    }

    // Change fullscreen video mode?
    if( _glfwWin.Fullscreen && mode != _glfwWin.ModeID )
    {
        // Change video mode
        _glfwSetVideoModeMODE( mode );

        // Clear the front buffer to black (avoid ugly desktop remains in
        // our OpenGL window)
        glGetIntegerv( GL_DRAW_BUFFER, &drawbuffer );
        glGetFloatv( GL_COLOR_CLEAR_VALUE, clearcolor );
        glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
        glClear( GL_COLOR_BUFFER_BIT );
        if( drawbuffer == GL_BACK )
        {
            _glfw_SwapBuffers( _glfwWin.DC );
        }
        glClearColor( clearcolor[0], clearcolor[1], clearcolor[2],
                      clearcolor[3] );
    }

    // Set window size (if not already changed)
    if( !sizechanged )
    {
        SetWindowPos( _glfwWin.Wnd, HWND_TOP, 0, 0, width, height,
                      SWP_NOOWNERZORDER | SWP_NOMOVE | SWP_NOZORDER );
    }
}


//========================================================================
// glfwSetWindowPos() - Set the window position.
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwSetWindowPos( int x, int y )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened || _glfwWin.Fullscreen )
    {
        return;
    }

    // Set window position
    (void) SetWindowPos( _glfwWin.Wnd, HWND_TOP, x, y, 0, 0,
                         SWP_NOOWNERZORDER | SWP_NOSIZE | SWP_NOZORDER );
}


//========================================================================
// glfwIconfyWindow() - Window iconification
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwIconifyWindow( void )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened || _glfwWin.Iconified )
    {
        return;
    }

    // Iconify window
    CloseWindow( _glfwWin.Wnd );

    // Window is now iconified
    _glfwWin.Iconified = GL_TRUE;

    // If we are in fullscreen mode we need to change video modes
    if( _glfwWin.Fullscreen )
    {
        // Change display settings to the desktop resolution
        ChangeDisplaySettings( NULL, 0 );
    }

    // Unlock mouse
    if( !_glfwWin.OldMouseLockValid )
    {
        _glfwWin.OldMouseLock = _glfwWin.MouseLock;
        _glfwWin.OldMouseLockValid = GL_TRUE;
        glfwEnable( GLFW_MOUSE_CURSOR );
    }
}


//========================================================================
// glfwRestoreWindow() - Window un-iconification
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwRestoreWindow( void )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened || !_glfwWin.Iconified )
    {
        return;
    }

    // If we are in fullscreen mode we need to change video modes
    if( _glfwWin.Fullscreen )
    {
        // Change display settings to the user selected mode
        _glfwSetVideoModeMODE( _glfwWin.ModeID );
    }

    // Un-iconify window
    OpenIcon( _glfwWin.Wnd );

    // Make sure that our window ends up on top of things
    ShowWindow( _glfwWin.Wnd, SW_SHOW );
    _glfwSetForegroundWindow( _glfwWin.Wnd );
    SetFocus( _glfwWin.Wnd );

    // Window is no longer iconified
    _glfwWin.Iconified = GL_FALSE;

    // Lock mouse, if necessary
    if( _glfwWin.OldMouseLockValid && _glfwWin.OldMouseLock )
    {
        glfwDisable( GLFW_MOUSE_CURSOR );
    }
    _glfwWin.OldMouseLockValid = GL_FALSE;
}


//========================================================================
// glfwSwapBuffers() - Swap buffers (double-buffering) and poll any new
// events.
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwSwapBuffers( void )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Check for window messages
    glfwPollEvents();

    // Update display-buffer
    if( _glfwWin.Opened )
    {
        _glfw_SwapBuffers( _glfwWin.DC );
    }
}


//========================================================================
// glfwSwapInterval() - Set double buffering swap interval (0 = vsync off)
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwSwapInterval( int interval )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    if( _wglSwapIntervalEXT )
    {
        _wglSwapIntervalEXT( interval );
    }
}


//========================================================================
// glfwGetWindowParam()
//========================================================================

GLFWAPI int GLFWAPIENTRY glfwGetWindowParam( int param )
{
    PIXELFORMATDESCRIPTOR pfd;
    DEVMODE dm;
    int     iPixelFormat, value, success, mode;

    // Is GLFW initialized?
    if( !_glfwInitialised )
    {
        return 0;
    }

    // Is the window opened?
    if( !_glfwWin.Opened )
    {
        if( param == GLFW_OPENED )
        {
            return GL_FALSE;
        }
        return 0;
    }

    // Obtain a detailed description of current pixel format
    if( (param >= GLFW_ACCELERATED && param <= GLFW_STENCIL_BITS) ||
        (param >= GLFW_ACCUM_RED_BITS && param <= GLFW_STEREO) )
    {
        iPixelFormat = _glfw_GetPixelFormat( _glfwWin.DC );
        _glfw_DescribePixelFormat( _glfwWin.DC, iPixelFormat,
                                   sizeof(PIXELFORMATDESCRIPTOR), &pfd );
    }

    // Get window parameter
    switch( param )
    {
    case GLFW_OPENED:
        value = GL_TRUE;
        break;
    case GLFW_ACTIVE:
        value = _glfwWin.Active;
        break;
    case GLFW_ICONIFIED:
        value = _glfwWin.Iconified;
        break;
    case GLFW_ACCELERATED:
        value =  (pfd.dwFlags & PFD_GENERIC_ACCELERATED) ||
                 !(pfd.dwFlags & PFD_GENERIC_FORMAT) ? 1 : 0;
        break;
    case GLFW_RED_BITS:
        value = pfd.cRedBits;
        break;
    case GLFW_GREEN_BITS:
        value = pfd.cGreenBits;
        break;
    case GLFW_BLUE_BITS:
        value = pfd.cBlueBits;
        break;
    case GLFW_ALPHA_BITS:
        value = pfd.cAlphaBits;
        break;
    case GLFW_DEPTH_BITS:
        value = pfd.cDepthBits;
        break;
    case GLFW_STENCIL_BITS:
        value = pfd.cStencilBits;
        break;
    case GLFW_ACCUM_RED_BITS:
        value = pfd.cAccumRedBits;
        break;
    case GLFW_ACCUM_GREEN_BITS:
        value = pfd.cAccumGreenBits;
        break;
    case GLFW_ACCUM_BLUE_BITS:
        value = pfd.cAccumBlueBits;
        break;
    case GLFW_ACCUM_ALPHA_BITS:
        value = pfd.cAccumAlphaBits;
        break;
    case GLFW_AUX_BUFFERS:
        value = pfd.cAuxBuffers;
        break;
    case GLFW_STEREO:
        value = pfd.dwFlags & PFD_STEREO ? GL_TRUE : GL_FALSE;
        break;
    case GLFW_REFRESH_RATE:
        mode = _glfwWin.Fullscreen ? _glfwWin.ModeID :
                                     ENUM_CURRENT_SETTINGS;
        dm.dmSize = sizeof( DEVMODE );
        success = EnumDisplaySettings( NULL, mode, &dm );
        if( success )
        {
            value = dm.dmDisplayFrequency;
            if( value <= 1 )
            {
                value = 0;
            }
        }
        else
        {
            value = 0;
        }
        break;
    default:
        value = 0;
        break;
    }

    return value;
}


//========================================================================
// glfwSetWindowSizeCallback() - Set callback function for window size
// changes
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwSetWindowSizeCallback( GLFWwindowsizefun cbfun )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Use default callback?
    _glfwWindowSizeCallback = cbfun;

    // Call the callback function to let the application know the current
    // window size
    if( _glfwWindowSizeCallback )
    {
        _glfwWindowSizeCallback( _glfwWin.Width, _glfwWin.Height );
    }
}


//========================================================================
// glfwPollEvents() - Poll for new window and input events
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwPollEvents( void )
{
    MSG msg;
    int winclosed = GL_FALSE;

    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Flag: mouse was not moved (will be changed by _glfwGetNextEvent if
    // there was a mouse move event)
    _glfwInput.MouseMoved = GL_FALSE;
    if( _glfwWin.MouseLock )
    {
        _glfwInput.OldMouseX = _glfwWin.Width/2;
        _glfwInput.OldMouseY = _glfwWin.Height/2;
    }
    else
    {
        _glfwInput.OldMouseX = _glfwInput.MousePosX;
        _glfwInput.OldMouseY = _glfwInput.MousePosY;
    }

    // Check for new window messages
    while( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) )
    {
        switch( msg.message )
        {
            // QUIT-message (from close window)?
            case WM_QUIT:
                winclosed = GL_TRUE;
                break;

            // Ok, send it to the window message handler
            default:
                DispatchMessage( &msg );
                break;
        }
    }

    // LSHIFT/RSHIFT fixup (keys tend to "stick" without this fix)
    // This is the only async event handling in GLFW, but it solves some
    // nasty problems.
    // Caveat: Does not work under Win 9x/ME.
    if( _glfwSys.WinVer >= _GLFW_WIN_NT4 )
    {
        int lshift_down, rshift_down;

        // Get current state of left and right shift keys
        lshift_down = (GetAsyncKeyState( VK_LSHIFT ) >> 15) & 1;
        rshift_down = (GetAsyncKeyState( VK_RSHIFT ) >> 15) & 1;

        // See if this differs from our belief of what has happened
        // (we only have to check for lost key up events)
        if( !lshift_down && _glfwInput.Key[ GLFW_KEY_LSHIFT ] == 1 )
        {
            _glfwInputKey( GLFW_KEY_LSHIFT, GLFW_RELEASE );
        }
        if( !rshift_down && _glfwInput.Key[ GLFW_KEY_RSHIFT ] == 1 )
        {
            _glfwInputKey( GLFW_KEY_RSHIFT, GLFW_RELEASE );
        }
    }

    // Did we have mouse movement in locked cursor mode?
    if( _glfwInput.MouseMoved && _glfwWin.MouseLock )
    {
        POINT pos;

        // Convert mouse coordinates to screen coordinates
        pos.x = _glfwWin.Width>>1;
        pos.y = _glfwWin.Height>>1;
        ClientToScreen( _glfwWin.Wnd, &pos );

        // Move the mouse pointer back to the window centre so that it
        // does not wander off...
        SetCursorPos( pos.x, pos.y );
    }

    // Was there a window close request?
    if( winclosed )
    {
        glfwCloseWindow();
    }
}
